/*

2clseq. Copyright Fahiem Bacchus 2000--2002

 A DPLL solver based on performing extensive reasoning with the binary
 clause subtheory. Pronounced "Two" "Clause" "Eq"

Version 2.0.

The software available at this site is copyright (c) 2000-2002 Fahiem
Bacchus. All right are reserved. Use of this software is permitted for
non-commercial research purposes, and it may be copied only for that
use. All copies must include this copyright message. This software and
any documentation and/or information supplied with it is distributed
on an as is basis. Fahiem Bacchus and the University of Toronto make
no warranties, express or implied, including but not limited to
implied warranties of merchantability and fitness for a particular
purpose, regarding the documentation, functions or performance of such
software, documentation and/or information.

@misc{Bacchus:2clseq,
	author  = { Fahiem Bacchus },
	year    = { 2002 },
	title   = { {2clseq} a {DPLL} solver employing extensive binary clause reasoning },
	note    = { Available from 
                    http://www.cs.toronto.edu/\~{}fbacchus/2clseq.html }
}

*/

#ifndef __resolver_H
#define __resolver_H

#include "binaryClauses.h"
#include "naryClauses.h"
#include "uniqueset.h"
#include "stats.h"
#include "dag.h"
#include "nogood.h"

extern bool DOEQS;
extern bool DOHYP;
extern bool DOLPB;

void DoPruneBacks(NoGood *ng, int lt, int btlevel);

/*******************************************************************
The resolution
******************************************************************/
typedef struct bcls {
  bcls *pnext;
  int l1,l2;
  DAG *dagnode;
} bcls;

typedef struct EQS {
  Lit* L;
  DAG* dagpos;
  DAG* dagneg;
} EQS; 

extern bcls *pbclsalloc;
extern bcls *pbclstack;
extern bcls *pclschkstack;
extern UniqueSet *clsToChk;
extern UniqueSet *upstack;
extern UniqueSet *litstochk;
extern UniqueSet *litbuf;
extern Lit** lits;
extern EQS* eqbuf;

extern int MRFailedLit; 
extern int HIGHESTBTLEVEL;
extern int DEEPESTLEVEL;

extern int sizebclsstack;
extern int sizeclschkstack;


const int BCLSALLOCFACTOR=20;

void Makebclstore();
void findallhits(Clause *c);

inline bcls* allocbcls() {
  //allocate a new bcls record
  if(!pbclsalloc)
    Makebclstore();
  bcls *pbcls = pbclsalloc;
  pbclsalloc=pbcls->pnext;
  return(pbcls);
}

inline void deallocbcls(bcls *pbcls) {
  //deallocate it.
  pbcls->pnext = pbclsalloc;
  pbclsalloc = pbcls;
}

/*We use two stacks a 2-clause stack and a unit prop stack.
  These stacks are used to store pending 2-clauses to be resolved, and
  literals to be unit propagated. Processing of UPs is always done
  prior to 2-clauses. So some 2-clauses	in the stack might become
  redundant prior to being resolved.

  Also a contradition might be discovered at any time. In which case
  entries on both stacks become stale.
 
  finally, it is possible for a lit to become equivalent to another, so
  pop lits that have been reduced by equivalences.
*/

inline void pushupstack(int l, DAG *dagnode) {
  switch(val(l)) {
  case F: 
    setContradiction(dagnode,getLitdagnode(negate(l)));
    break;
  case T:
    break;
  case U:
    if(upstack->member(negate(l))) {
      setContradiction(dagnode,getLitdagnode(negate(l)));
    }
    else if(!upstack->member(l)) {
      setLitdagnode(l,dagnode);
      markundoLitdagnode(l);
      upstack->pushnew(l);
      if(DOLPB && dagnode->type==NORMAL) {
	NoGood *ng=allocNoGood(LEVEL+1);
	ng->unionFromDAG(dagnode);
	DoPruneBacks(ng,negate(l),LEVEL);
	deallocNoGood(ng);
      }
    }
  }
}

inline bool emptyupstack() {
  return(upstack->size==0);
}

inline DAG* computeLitEqdagnode(int lt);

inline int popupstack() {
  int lt=upstack->pop();
  int eqlt=eqlit(lt);
  if(eqlt != lt && !getLitdagnode(eqlt)) {
	//printf("=>popupstack reducing eqlit!\n");
	setLitdagnode(eqlt,allocDAGNODE_NORMAL(getLitdagnode(lt),
					       computeLitEqdagnode(lt)));
	markundoLitdagnode(lt);
  }
  return(eqlt);
}

inline void pushbclsstack(int l1, int l2, DAG *dagnode) {
  //check bcls and push it on the bcls stack to be resolved
  //if appropriate.
#ifdef _TRACE
  printf("=>pushbclsstack(%d[%c],%d[%c]).\n",to_ff(l1),
		 ff_tval(val(l1)), to_ff(l2), ff_tval(val(l2)));
  //  if(HvalhaveBclause(l1,l2,RESOLVED))
	//    
  if(haveBclause(l1,l2))
	printf("HaveBclause.\n");
	//  if(HvalhaveBclause(l1,l2,INSTACK))
	//    printf("HaveBclause with INSTACK\n");
#endif
  //bucket *pb = intTolit(l1)->litHtable;
  //bucket *notpb = intTolit(negate(l1))->litHtable;
  DAG *dnode1;

  //  if(haveBclause(pb,l2))
  if(haveBclause(l1,l2))
    return;
  else if(isFalse(l1)) {
    pushupstack(l2,allocDAGNODE_NORMAL(dagnode,getLitdagnode(negate(l1)))); 
  }
  //  else if(DNodehaveBclause(notpb,l2,dnode1)) {
  else if(DNodehaveBclause(negate(l1),l2,dnode1)) {
    pushupstack(l2,allocDAGNODE_NORMAL(dagnode,dnode1));
  }
  else if(isFalse(l2)) {
    pushupstack(l1,allocDAGNODE_NORMAL(dagnode,getLitdagnode(negate(l2))));
  }
  //  else if(DNodehaveBclause(pb,negate(l2),dnode1)) {
  else if(DNodehaveBclause(l1,negate(l2),dnode1)) {
    pushupstack(l1,allocDAGNODE_NORMAL(dagnode,dnode1));
  }
  else {
    //put on to be resolved stack
#ifdef _TRACE
    printf("==pushbclsstack(%d,%d) added to stack.\n",to_ff(l1),to_ff(l2));
#endif
	sizebclsstack++;
	bcls *pbcls=allocbcls();
	pbcls->l1=l1;
	pbcls->l2=l2;
	pbcls->dagnode=dagnode;
	pbcls->pnext=pbclstack;
	pbclstack=pbcls;
	//markBclause(l1,l2,INSTACK,dagnode);
  }
}

inline void pushbclsstack(bcls *pbcls) {
  int l1=pbcls->l1;
  int l2=pbcls->l2;
  //check bcls and push it on the bcls stack to be resolved
  //if appropriate.
#ifdef _TRACE
  printf("=>pushbclsstack(%d[%c],%d[%c]).\n",to_ff(l1),
		 ff_tval(val(l1)), to_ff(l2), ff_tval(val(l2)));
  //  if(HvalhaveBclause(l1,l2,RESOLVED))
  //    printf("HaveBclause with RESOLVED\n");
  if(haveBclause(l1,l2))
	printf("HaveBclause.\n");
  //  if(HvalhaveBclause(l1,l2,INSTACK))
  //    printf("HaveBclause with INSTACK\n");
#endif
  //bucket *pb = intTolit(l1)->litHtable;
  //bucket *notpb = intTolit(negate(l1))->litHtable;
  DAG *dnode1;
	
  //  if(haveBclause(pb,l2))
  if(haveBclause(l1,l2))
    return;

  else if(isFalse(l1)) {
    pushupstack(l2,allocDAGNODE_NORMAL(pbcls->dagnode,getLitdagnode(negate(l1))));
  }
  //  else if(DNodehaveBclause(notpb,l2,dnode1)) {
  else if(DNodehaveBclause(negate(l1),l2,dnode1)) {
    pushupstack(l2,allocDAGNODE_NORMAL(pbcls->dagnode,dnode1));
  }
  else if(isFalse(l2)) {
    pushupstack(l1,allocDAGNODE_NORMAL(pbcls->dagnode,getLitdagnode(negate(l2))));
  }
  //  else if(DNodehaveBclause(pb,negate(l2),dnode1)) {
  else if(DNodehaveBclause(l1,negate(l2),dnode1)) {
    pushupstack(l1,allocDAGNODE_NORMAL(pbcls->dagnode,dnode1));
  }

  else {
    //put on to be resolved stack
#ifdef _TRACE
    printf("==pushbclsstack(%d,%d) added to stack.\n",to_ff(l1),to_ff(l2));
#endif
	sizebclsstack++;
	pbcls->pnext=pbclstack;
	pbclstack=pbcls;
	//markBclause(l1,l2,INSTACK,pbcls->dagnode);
  }
}

inline bool emptybclstack() {
  return(pbclstack==NULL);
}

inline bcls* popbclsstack() {
  sizebclsstack--;
  bcls* pbcls=pbclstack;
  pbclstack=pbcls->pnext;
  //if(HvalhaveBclause(pbcls->l1,pbcls->l2,INSTACK))
  //unmarkBclause(pbcls->l1,pbcls->l2);
  DAG *dnode=pbcls->dagnode;
  int eql1=eqlit(pbcls->l1);
  if(eql1!=pbcls->l1) {
	//printf("=>popbclsstack reducing eqlit!\n");
	dnode=allocDAGNODE_NORMAL(dnode,
							  computeLitEqdagnode(pbcls->l1));
  }
  int eql2=eqlit(pbcls->l2); 
  if(eql1!=pbcls->l2) {
	//printf("=>popbclsstack reducing eqlit!\n");
	dnode=allocDAGNODE_NORMAL(dnode,
							  computeLitEqdagnode(pbcls->l2));
  }
  pbcls->l1=eql1;
  pbcls->l2=eql2;
  pbcls->dagnode=dnode;
  return(pbcls);
}

inline void pushclschkstack(int l1, int l2) {
  //Push a bclause to be check later for clause hits.
#ifdef _TRACE
  printf("=>pushclschkstack(%d[%c],%d[%c]).\n",to_ff(l1),
		 ff_tval(val(l1)), to_ff(l2), ff_tval(val(l2)));
#endif
  sizeclschkstack++;
  bcls *pbcls=allocbcls();
  pbcls->l1=l1;
  pbcls->l2=l2;
  pbcls->pnext=pclschkstack;
  pclschkstack=pbcls;
}

inline void pushclschkstack(Lit* l1, Lit* l2) {
  //Push a bclause to be check later for clause hits.
#ifdef _TRACE
  printf("=>pushclschkstack(%d[%c],%d[%c]).\n",to_ff(l1->num),
		 ff_tval(val(l1)), to_ff(l2->num), ff_tval(val(l2)));
#endif
  sizeclschkstack++;
  bcls *pbcls=allocbcls();
  pbcls->l1=l1->num;
  pbcls->l2=l2->num;
  pbcls->pnext=pclschkstack;
  pclschkstack=pbcls;
}

inline bool emptyclschkstack() {
  return(pclschkstack==NULL);
}

inline bcls* popclschkstack() {
  sizeclschkstack--;
  bcls* pbcls=pclschkstack;
  pclschkstack=pbcls->pnext;
  return(pbcls);
}

inline void emptystacks() {
  //On contradiction empty both stacks.
  upstack->empty();
  while(!emptybclstack())
    deallocbcls(popbclsstack());
  while(!emptyclschkstack())
    deallocbcls(popclschkstack());
  clsToChk->empty();
}

void UnitProp(int l);
void resolvebcls(bcls *pbcls);
void resolveEqcls(int l1, int l2);
void checkClauseHits(bcls *pbcls);
void checkClauseHits(Lit *l1, Lit *l2);

void inline convertToBinaryClause(Clause* pcls) {
  int l1=pcls->start.pnext->plit->num;
  int l2=pcls->start.pnext->pnext->plit->num;
  pushbclsstack(l1,l2,pcls->dagnode);
  markundoClauseTrue(pcls);
  delinkNClause(pcls);
}

bool inline parent(Lit *pl) {
  //True if pl has a parent. I.e., another lit 
  //that pl passes on all of its 2-clauses to
  Lit *l=negate(pl);
  return(l->twoClauses.pnext->pnext==NULL);
}

bool inline parent(int il) {
  //True if l has a parent. I.e., another lit 
  //that l passes on all of its 2-clauses to
  Lit *l=THELITS+negate(il);
  return(l->twoClauses.pnext->pnext==NULL);
}

void inline propagateStacks() {
  //do all stacked up unit props and binary clause resolutions.
  while(!CONTRADICTION.val &&
		(!emptyupstack() || !emptybclstack() || !emptyclschkstack()
		 || !clsToChk->isEmpty())) {
    if(!emptyupstack()) 
      UnitProp(popupstack());
    else if(!emptybclstack()) 
      resolvebcls(popbclsstack());
    else if(!emptyclschkstack())
	  checkClauseHits(popclschkstack());
    else if(!clsToChk->isEmpty())
	  findallhits(THECLAUSES+clsToChk->pop());
  }
  if(CONTRADICTION.val)
    emptystacks();
}


bool inline checkforNewbClause(Clause *c, Lit* L, int &missedlit, int &missedcount) {
  //check if l generates a new bclause from clause c
  missedcount=0;
  //bucket *pb = L->litHtable;
  for(ClsLit *cl=c->start.pnext; cl->pnext; cl=cl->pnext) {
	//    if(haveBclause(pb,cl->plit->num)) {
	if(haveBclause(L,cl->plit->num)) {
      return false;
    }
	//    if(!haveBclause(pb,negate(cl->plit->num))) {
	if(!haveBclause(L,negate(cl->plit->num))) {
      missedcount++;
      if(missedcount>1)
		return false;
      missedlit=cl->plit->num;
    }
  }
  return true;
}

inline void checkClauseHits(bcls *pbcls) {
  Lit* l1=eqlit(intTolit(pbcls->l1));
  Lit* l2=eqlit(intTolit(pbcls->l2));
  deallocbcls(pbcls);
  checkClauseHits(l1, l2);
}

inline DAG* computeDAGnode(Lit *L, Clause *c) {
  //compute a dag for L generating a new binary of unary clause from
  //hitting clause c
  DAG *dnode1, *dnode2;
  dnode1=c->dagnode;

  //bucket *pb = L->litHtable;
  for(ClsLit *cl=c->start.pnext; cl->pnext; cl=cl->pnext) 
	//    if(DNodehaveBclause(pb,negate(cl->plit->num),dnode2))
	if(DNodehaveBclause(L,negate(cl->plit->num),dnode2)) 
      dnode1=allocDAGNODE_NORMAL(dnode2,dnode1);
  return dnode1;
}

inline DAG* computeLitEqdagnode(int lt) {
  //compute a dag for lit lt being equivalent to its currently
  //equivalent literal. This follows a chain of equivalences.
  Lit *Lt=intTolit(lt);
  DAG *dnode1=Lt->dagEqnode;
  Lt=Lt->equivlit;
  while(Lt->equivlit!=Lt) {
	//printf("computeLitEqdagnode following chain\n");
	dnode1=allocDAGNODE_NORMAL(Lt->dagEqnode,dnode1);
	Lt=Lt->equivlit;
  }
  return(dnode1);
}

inline void checkClauseHits(Lit *l1, Lit *l2) {
  LitCls *nclauses;
  int missedlit, missedcount;

#ifdef _TRACE
  printf("=>CNCHs(%d,%d)\n", to_ff(l1->num),to_ff(l2->num));
  printLit(l1);
  printLit(negate(l1));
  printLit(l2);
  printLit(negate(l2));
#endif _TRACE
	
  if(isInactive(l1) || isInactive(l2))
    return;

  bool foundAhit=false;
  if(numBcls(l1) > 1) {
    litHitsChecked++;
    //if(parent(l1)) parentHitsChecked++;
    for(nclauses=(negate(l1))->nClauses.pnext;nclauses->pnext;nclauses=nclauses->pnext) {
      clauseHitsChecked++;
      //if(parent(l1)) clauseHitsChecked++;
      if(checkforNewbClause(nclauses->pcls,l2,missedlit,missedcount)) {
		if(!foundAhit) {
		  //if(parent(l1)) parentHitsFound++;
		  litHitsFound++;
		  foundAhit=true;
		}
		clauseHitsFound++;
		//if(parent(l1)) clauseHitsFound++;
		if(missedcount==0) {
		  pushupstack(l2->num,computeDAGnode(l2,nclauses->pcls));
		  return;
		}
		else 
		  pushbclsstack(l2->num,missedlit,computeDAGnode(l2,nclauses->pcls));
      }
    }
  }
	
  if(numBcls(l1) > 1) {
    foundAhit=false;
    litHitsChecked++;
    //if(parent(l2)) parentHitsChecked++;
    for(nclauses=(negate(l2))->nClauses.pnext;nclauses->pnext;
		nclauses=nclauses->pnext) {
      clauseHitsChecked++;
      //if(parent(l2)) clauseHitsChecked++;
      if(checkforNewbClause(nclauses->pcls,l1,missedlit,missedcount)) {
		if(!foundAhit) {
		  //if(parent(l2)) parentHitsFound++;
		  litHitsFound++;
		  foundAhit=true;
		}
		clauseHitsFound++;
		//if(parent(l2)) clauseHitsFound++;
		if(missedcount==0) {
		  pushupstack(l1->num,computeDAGnode(l1,nclauses->pcls));
		  return;
		}
		else 
		  pushbclsstack(l1->num,missedlit,computeDAGnode(l1,nclauses->pcls));
      }
    }
  }
}

inline void findallhits(Clause *c) {
  //Find all literals hitting clause
  if(!c->active)
    return;
	
  nClausesChecked++;
  bool foundhit=false;
	
  int min1=NLITS;
  int min2=NLITS;
  Lit *L1=NULL;
  Lit *L2=NULL;

  //find the two lits in the clause whose negations have min binary clauses.
  for(ClsLit *pcl=c->start.pnext;pcl->pnext;pcl=pcl->pnext) {
    Lit *L=negate(pcl->plit); 
    if(numBcls(L)<min1) {
      min2=min1;
      L2=L1;
      min1=numBcls(L);
      L1=L;
    }
    else if(numBcls(L)<min2) {
      min2=numBcls(L);
      L2=L;
    }
  }

  //Now check the union of their bclauses
  int missedlit, missedcount;
  BinLitCls *pb;
  litbuf->empty();
  for(pb=&(L1->twoClauses); pb->pnext; pb=pb->pnext) {
    litbuf->pushnew(pb->otherlit->num);
    if(checkforNewbClause(c, pb->otherlit, missedlit, missedcount)) {
      if(!foundhit) {
		nClausesHits++;
		foundhit=true;
      }
      if(missedcount == 0) 
		pushupstack(pb->otherlit->num,computeDAGnode(pb->otherlit,c));
      else 
		pushbclsstack(pb->otherlit->num,missedlit,computeDAGnode(pb->otherlit,c));
    }
  }
  for(pb=&(L2->twoClauses); pb->pnext; pb=pb->pnext) {
    if(!litbuf->member(pb->otherlit->num)
       && checkforNewbClause(c, pb->otherlit, missedlit, missedcount)) {
      if(!foundhit) {
		nClausesHits++;
		foundhit=true;
      }
      if(missedcount == 0) 
		pushupstack(pb->otherlit->num,computeDAGnode(pb->otherlit,c));
      else 
		pushbclsstack(pb->otherlit->num,missedlit,computeDAGnode(pb->otherlit,c));
    }
  }
  litbuf->empty();
}

#endif /* __resolver_H */
